"""Generated implementation of secrets_config."""

# WARNING DO NOT EDIT
# This code was generated from secrets-config.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass


@dataclasses.dataclass(frozen=True)
class SecretsConfig(abc.ABC):
    """Configuration for a secret configuration item."""
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    @classmethod
    def json_schema(cls) -> SecretsConfig:
        """JSON schema for variant SecretsConfig.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> SecretsConfig:
        """Validate and parse JSON data into an instance of SecretsConfig.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of SecretsConfig.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing SecretsConfig", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class BasicSecretsConfig(SecretsConfig):
    """A secret passed directly.
    
    Args:
        secret (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "basic"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    secret: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for BasicSecretsConfig data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "secret": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "secret",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> BasicSecretsConfig:
        """Validate and parse JSON data into an instance of BasicSecretsConfig.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of BasicSecretsConfig.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return BasicSecretsConfig(
                secret=str(data["secret"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing BasicSecretsConfig",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "secret": str(self.secret)
        }


@dataclasses.dataclass(frozen=True)
class AWSSMSecretsConfig(SecretsConfig):
    """A secret stored in AWS Secret Manager.
    
    Args:
        secretId (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "awssm"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    secretId: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for AWSSMSecretsConfig data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "secretId": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "secretId",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> AWSSMSecretsConfig:
        """Validate and parse JSON data into an instance of AWSSMSecretsConfig.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of AWSSMSecretsConfig.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return AWSSMSecretsConfig(
                secretId=str(data["secretId"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing AWSSMSecretsConfig",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "secretId": str(self.secretId)
        }


@dataclasses.dataclass(frozen=True)
class GCPSMSecretsConfig(SecretsConfig):
    """A secret stored in GCP Secret Manager.
    
    Args:
        secretProject (str): A data field.
        secretId (str): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "gcpsm"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    secretProject: str
    secretId: str
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for GCPSMSecretsConfig data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "secretProject": {
                    "type": "string"
                },
                "secretId": {
                    "type": "string"
                }
            },
            "required": [
                "adt_type",
                "secretProject",
                "secretId",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> GCPSMSecretsConfig:
        """Validate and parse JSON data into an instance of GCPSMSecretsConfig.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of GCPSMSecretsConfig.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return GCPSMSecretsConfig(
                secretProject=str(data["secretProject"]),
                secretId=str(data["secretId"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing GCPSMSecretsConfig",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "secretProject": str(self.secretProject),
            "secretId": str(self.secretId)
        }



"""Generated implementation of user_group."""

# WARNING DO NOT EDIT
# This code was generated from user-group.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..roles import Role
from ..user import UserGroupId, UserId


@dataclasses.dataclass(frozen=True)
class UserGroupVersionId:
    """Unique identifier a specific version of a user group.
    
    Args:
        value (uuid.UUID): A data field.
    """
    
    value: uuid.UUID
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserGroupVersionId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string",
            "format": "uuid"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserGroupVersionId:
        """Validate and parse JSON data into an instance of UserGroupVersionId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserGroupVersionId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserGroupVersionId(uuid.UUID(hex=data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing UserGroupVersionId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> UserGroupVersionId:
        """Parse a JSON string such as a dictionary key."""
        return UserGroupVersionId((lambda s: uuid.UUID(hex=s))(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class UserGroupName:
    """Unique name of a user group.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserGroupName data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserGroupName:
        """Validate and parse JSON data into an instance of UserGroupName.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserGroupName.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserGroupName(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing UserGroupName", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> UserGroupName:
        """Parse a JSON string such as a dictionary key."""
        return UserGroupName(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ExternalGroupId:
    """Unique identifier for external identity provider.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ExternalGroupId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ExternalGroupId:
        """Validate and parse JSON data into an instance of ExternalGroupId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ExternalGroupId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ExternalGroupId(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ExternalGroupId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> ExternalGroupId:
        """Parse a JSON string such as a dictionary key."""
        return ExternalGroupId(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


class UserGroupMemberSource(enum.Enum):
    """User group member source type."""
    Anaml = "anaml"
    """Created internally via Anaml API."""
    External = "external"
    """Imported from external identity provider."""
    
    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'UserGroupMemberSource'.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [
                        "anaml",
                        "external",
                    ]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserGroupMemberSource:
        """Validate and parse JSON data into an instance of UserGroupMemberSource.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserGroupMemberSource.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserGroupMemberSource(str(data['adt_type']))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing UserGroupMemberSource", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            JSON data ready to be serialised.
        """
        return {'adt_type': self.value}
    
    @classmethod
    def from_json_key(cls, data: str) -> UserGroupMemberSource:
        """Validate and parse a value from a JSON dictionary key.
        
        Args:
            data (str): JSON data to validate and parse.
        
        Returns:
            An instance of UserGroupMemberSource.
        """
        return UserGroupMemberSource(str(data))
    
    def to_json_key(self) -> str:
        """Serialised this instanse as a JSON string for use as a dictionary key.
        
        Returns:
            A JSON string ready to be used as a key.
        """
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class UserGroupMember:
    """User group member.
    
    Args:
        userId (UserId): A data field.
        source (UserGroupMemberSource): A data field.
    """
    
    userId: UserId
    source: UserGroupMemberSource
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserGroupMember data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "userId": UserId.json_schema(),
                "source": UserGroupMemberSource.json_schema()
            },
            "required": [
                "userId",
                "source",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserGroupMember:
        """Validate and parse JSON data into an instance of UserGroupMember.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserGroupMember.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserGroupMember(
                userId=UserId.from_json(data["userId"]),
                source=UserGroupMemberSource.from_json(data["source"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing UserGroupMember",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "userId": self.userId.to_json(),
            "source": self.source.to_json()
        }


@dataclasses.dataclass(frozen=True)
class UserGroup:
    """Group of Anaml users.
    
    Args:
        id (UserGroupId): A data field.
        version (UserGroupVersionId): A data field.
        name (UserGroupName): A data field.
        description (str): A data field.
        members (typing.List[UserGroupMember]): A data field.
        created (datetime.datetime): A data field.
        modified (datetime.datetime): A data field.
        roles (typing.List[Role]): A data field.
        predecessor (typing.Optional[UserGroupVersionId]): A data field.
        externalGroupId (typing.Optional[ExternalGroupId]): A data field.
    """
    
    id: UserGroupId
    version: UserGroupVersionId
    name: UserGroupName
    description: str
    members: typing.List[UserGroupMember]
    created: datetime.datetime
    modified: datetime.datetime
    roles: typing.List[Role]
    predecessor: typing.Optional[UserGroupVersionId]
    externalGroupId: typing.Optional[ExternalGroupId]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserGroup data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "id": UserGroupId.json_schema(),
                "version": UserGroupVersionId.json_schema(),
                "name": UserGroupName.json_schema(),
                "description": {
                    "type": "string"
                },
                "members": {
                    "type": "array",
                    "item": UserGroupMember.json_schema()
                },
                "created": {
                    "type": "string",
                    "format": "date-time"
                },
                "modified": {
                    "type": "string",
                    "format": "date-time"
                },
                "roles": {
                    "type": "array",
                    "item": Role.json_schema()
                },
                "predecessor": {
                    "oneOf": [
                        {"type": "null"},
                        UserGroupVersionId.json_schema(),
                    ]
                },
                "externalGroupId": {
                    "oneOf": [
                        {"type": "null"},
                        ExternalGroupId.json_schema(),
                    ]
                }
            },
            "required": [
                "id",
                "version",
                "name",
                "description",
                "members",
                "created",
                "modified",
                "roles",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserGroup:
        """Validate and parse JSON data into an instance of UserGroup.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserGroup.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserGroup(
                id=UserGroupId.from_json(data["id"]),
                version=UserGroupVersionId.from_json(data["version"]),
                name=UserGroupName.from_json(data["name"]),
                description=str(data["description"]),
                members=[UserGroupMember.from_json(v) for v in data["members"]],
                created=isodate.parse_datetime(data["created"]),
                modified=isodate.parse_datetime(data["modified"]),
                roles=[Role.from_json(v) for v in data["roles"]],
                predecessor=(
                    lambda v: v and UserGroupVersionId.from_json(v)
                )(
                    data.get("predecessor", None)
                ),
                externalGroupId=(
                    lambda v: v and ExternalGroupId.from_json(v)
                )(
                    data.get("externalGroupId", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing UserGroup",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "id": self.id.to_json(),
            "version": self.version.to_json(),
            "name": self.name.to_json(),
            "description": str(self.description),
            "members": [v.to_json() for v in self.members],
            "created": self.created.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "modified": self.modified.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "roles": [v.to_json() for v in self.roles],
            "predecessor": (lambda v: v and v.to_json())(self.predecessor),
            "externalGroupId": (lambda v: v and v.to_json())(self.externalGroupId)
        }


@dataclasses.dataclass(frozen=True)
class UserGroupCreationRequest:
    """Request to create a new user group.
    
    Args:
        name (UserGroupName): A data field.
        description (str): A data field.
        roles (typing.List[Role]): A data field.
        members (typing.List[UserGroupMember]): A data field.
        externalGroupId (typing.Optional[ExternalGroupId]): A data field.
    """
    
    name: UserGroupName
    description: str
    roles: typing.List[Role]
    members: typing.List[UserGroupMember]
    externalGroupId: typing.Optional[ExternalGroupId]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for UserGroupCreationRequest data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "name": UserGroupName.json_schema(),
                "description": {
                    "type": "string"
                },
                "roles": {
                    "type": "array",
                    "item": Role.json_schema()
                },
                "members": {
                    "type": "array",
                    "item": UserGroupMember.json_schema()
                },
                "externalGroupId": {
                    "oneOf": [
                        {"type": "null"},
                        ExternalGroupId.json_schema(),
                    ]
                }
            },
            "required": [
                "name",
                "description",
                "roles",
                "members",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> UserGroupCreationRequest:
        """Validate and parse JSON data into an instance of UserGroupCreationRequest.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of UserGroupCreationRequest.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return UserGroupCreationRequest(
                name=UserGroupName.from_json(data["name"]),
                description=str(data["description"]),
                roles=[Role.from_json(v) for v in data["roles"]],
                members=[UserGroupMember.from_json(v) for v in data["members"]],
                externalGroupId=(
                    lambda v: v and ExternalGroupId.from_json(v)
                )(
                    data.get("externalGroupId", None)
                ),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing UserGroupCreationRequest",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "name": self.name.to_json(),
            "description": str(self.description),
            "roles": [v.to_json() for v in self.roles],
            "members": [v.to_json() for v in self.members],
            "externalGroupId": (lambda v: v and v.to_json())(self.externalGroupId)
        }



"""Generated implementation of jobs."""

# WARNING DO NOT EDIT
# This code was generated from jobs.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass


@dataclasses.dataclass(frozen=True)
class FeatureStoreRunId:
    """Unique identifier of a feature store run.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FeatureStoreRunId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FeatureStoreRunId:
        """Validate and parse JSON data into an instance of FeatureStoreRunId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FeatureStoreRunId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FeatureStoreRunId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing FeatureStoreRunId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return self.value
    
    @classmethod
    def from_json_key(cls, data: str) -> FeatureStoreRunId:
        """Parse a JSON string such as a dictionary key."""
        return FeatureStoreRunId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class TableMonitoringJobId:
    """Unique identifier of a table monitoring job.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableMonitoringJobId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableMonitoringJobId:
        """Validate and parse JSON data into an instance of TableMonitoringJobId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableMonitoringJobId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableMonitoringJobId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing TableMonitoringJobId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return int(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> TableMonitoringJobId:
        """Parse a JSON string such as a dictionary key."""
        return TableMonitoringJobId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class TableMonitoringRunId:
    """Unique identifier of a table monitoring job run.
    
    Args:
        value (int): A data field.
    """
    
    value: int
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    def __int__(self) -> int:
        """Return an int of the wrapped value."""
        return int(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for TableMonitoringRunId data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "integer"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> TableMonitoringRunId:
        """Validate and parse JSON data into an instance of TableMonitoringRunId.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of TableMonitoringRunId.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return TableMonitoringRunId(int(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing TableMonitoringRunId", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return self.value
    
    @classmethod
    def from_json_key(cls, data: str) -> TableMonitoringRunId:
        """Parse a JSON string such as a dictionary key."""
        return TableMonitoringRunId(int(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


class RunStatus(enum.Enum):
    """Status of a job."""
    Pending = "pending"
    """The job is waiting to be started."""
    Running = "running"
    """The job is running."""
    Completed = "completed"
    """The job has completed successfully."""
    Cancel = "cancel"
    """A user has requested that the job should be cancelled."""
    Cancelled = "cancelled"
    """The job was cancelled by a user."""
    Failed = "failed"
    """The job has failed."""
    Redundant = "redundant"
    """The job was scheduled but no work is required of it."""
    
    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'RunStatus'.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [
                        "pending",
                        "running",
                        "completed",
                        "cancel",
                        "cancelled",
                        "failed",
                        "redundant",
                    ]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> RunStatus:
        """Validate and parse JSON data into an instance of RunStatus.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of RunStatus.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return RunStatus(str(data['adt_type']))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing RunStatus", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            JSON data ready to be serialised.
        """
        return {'adt_type': self.value}
    
    @classmethod
    def from_json_key(cls, data: str) -> RunStatus:
        """Validate and parse a value from a JSON dictionary key.
        
        Args:
            data (str): JSON data to validate and parse.
        
        Returns:
            An instance of RunStatus.
        """
        return RunStatus(str(data))
    
    def to_json_key(self) -> str:
        """Serialised this instanse as a JSON string for use as a dictionary key.
        
        Returns:
            A JSON string ready to be used as a key.
        """
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class RunError:
    """An error with a potential stack trace.
    
    Args:
        message (str): A data field.
        stackTrace (typing.Optional[str]): A data field.
    """
    
    message: str
    stackTrace: typing.Optional[str]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for RunError data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "message": {
                    "type": "string"
                },
                "stackTrace": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                }
            },
            "required": [
                "message",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> RunError:
        """Validate and parse JSON data into an instance of RunError.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of RunError.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return RunError(
                message=str(data["message"]),
                stackTrace=(lambda v: v and str(v))(data.get("stackTrace", None)),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing RunError",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "message": str(self.message),
            "stackTrace": (lambda v: v and str(v))(self.stackTrace)
        }



"""Generated implementation of reports."""

# WARNING DO NOT EDIT
# This code was generated from reports.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass


class ActionType(enum.Enum):
    """Actions to be reported."""
    Create = "create"
    Update = "update"
    Delete = "delete"
    
    @classmethod
    def json_schema(cls) -> dict:
        """JSON schema for 'ActionType'.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [
                        "create",
                        "update",
                        "delete",
                    ]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ActionType:
        """Validate and parse JSON data into an instance of ActionType.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ActionType.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ActionType(str(data['adt_type']))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing ActionType", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            JSON data ready to be serialised.
        """
        return {'adt_type': self.value}
    
    @classmethod
    def from_json_key(cls, data: str) -> ActionType:
        """Validate and parse a value from a JSON dictionary key.
        
        Args:
            data (str): JSON data to validate and parse.
        
        Returns:
            An instance of ActionType.
        """
        return ActionType(str(data))
    
    def to_json_key(self) -> str:
        """Serialised this instanse as a JSON string for use as a dictionary key.
        
        Returns:
            A JSON string ready to be used as a key.
        """
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class ReportAction:
    """Report of an action performed by a user.
    
    Reports are computed asynchronously after the fact.
    
    Args:
        timestamp (datetime.datetime): A data field.
        branch (str): A data field.
        user (str): A data field.
        action (ActionType): A data field.
        object_type (str): A data field.
        object_id (int): A data field.
    """
    
    timestamp: datetime.datetime
    branch: str
    user: str
    action: ActionType
    object_type: str
    object_id: int
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ReportAction data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "timestamp": {
                    "type": "string",
                    "format": "date-time"
                },
                "branch": {
                    "type": "string"
                },
                "user": {
                    "type": "string"
                },
                "action": ActionType.json_schema(),
                "object_type": {
                    "type": "string"
                },
                "object_id": {
                    "type": "integer"
                }
            },
            "required": [
                "timestamp",
                "branch",
                "user",
                "action",
                "object_type",
                "object_id",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ReportAction:
        """Validate and parse JSON data into an instance of ReportAction.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ReportAction.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ReportAction(
                timestamp=isodate.parse_datetime(data["timestamp"]),
                branch=str(data["branch"]),
                user=str(data["user"]),
                action=ActionType.from_json(data["action"]),
                object_type=str(data["object_type"]),
                object_id=int(data["object_id"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ReportAction",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "timestamp": self.timestamp.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "branch": str(self.branch),
            "user": str(self.user),
            "action": self.action.to_json(),
            "object_type": str(self.object_type),
            "object_id": int(self.object_id)
        }



"""Generated implementation of lineage."""

# WARNING DO NOT EDIT
# This code was generated from lineage.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass

from ..destination import Destination
from ..entity import Entity
from ..entity_mapping import EntityMapping
from ..entity_population import EntityPopulation
from ..event_store import EventStore
from ..feature import Feature
from ..feature_set import FeatureSet
from ..feature_store import FeatureStore
from ..source import Source
from ..table import Table


@dataclasses.dataclass(frozen=True)
class Lineage:
    """Representation of data lineage.
    
    Args:
        eventStores (typing.List[EventStore]): A data field.
        sources (typing.List[Source]): A data field.
        tables (typing.List[Table]): A data field.
        entities (typing.List[Entity]): A data field.
        mappings (typing.List[EntityMapping]): A data field.
        populations (typing.List[EntityPopulation]): A data field.
        features (typing.List[Feature]): A data field.
        featureSets (typing.List[FeatureSet]): A data field.
        featureStores (typing.List[FeatureStore]): A data field.
        destinations (typing.List[Destination]): A data field.
    """
    
    eventStores: typing.List[EventStore]
    sources: typing.List[Source]
    tables: typing.List[Table]
    entities: typing.List[Entity]
    mappings: typing.List[EntityMapping]
    populations: typing.List[EntityPopulation]
    features: typing.List[Feature]
    featureSets: typing.List[FeatureSet]
    featureStores: typing.List[FeatureStore]
    destinations: typing.List[Destination]
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Lineage data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "eventStores": {
                    "type": "array",
                    "item": EventStore.json_schema()
                },
                "sources": {
                    "type": "array",
                    "item": Source.json_schema()
                },
                "tables": {
                    "type": "array",
                    "item": Table.json_schema()
                },
                "entities": {
                    "type": "array",
                    "item": Entity.json_schema()
                },
                "mappings": {
                    "type": "array",
                    "item": EntityMapping.json_schema()
                },
                "populations": {
                    "type": "array",
                    "item": EntityPopulation.json_schema()
                },
                "features": {
                    "type": "array",
                    "item": Feature.json_schema()
                },
                "featureSets": {
                    "type": "array",
                    "item": FeatureSet.json_schema()
                },
                "featureStores": {
                    "type": "array",
                    "item": FeatureStore.json_schema()
                },
                "destinations": {
                    "type": "array",
                    "item": Destination.json_schema()
                }
            },
            "required": [
                "eventStores",
                "sources",
                "tables",
                "entities",
                "mappings",
                "populations",
                "features",
                "featureSets",
                "featureStores",
                "destinations",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Lineage:
        """Validate and parse JSON data into an instance of Lineage.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Lineage.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Lineage(
                eventStores=[EventStore.from_json(v) for v in data["eventStores"]],
                sources=[Source.from_json(v) for v in data["sources"]],
                tables=[Table.from_json(v) for v in data["tables"]],
                entities=[Entity.from_json(v) for v in data["entities"]],
                mappings=[EntityMapping.from_json(v) for v in data["mappings"]],
                populations=[EntityPopulation.from_json(v) for v in data["populations"]],
                features=[Feature.from_json(v) for v in data["features"]],
                featureSets=[FeatureSet.from_json(v) for v in data["featureSets"]],
                featureStores=[FeatureStore.from_json(v) for v in data["featureStores"]],
                destinations=[Destination.from_json(v) for v in data["destinations"]],
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing Lineage",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "eventStores": [v.to_json() for v in self.eventStores],
            "sources": [v.to_json() for v in self.sources],
            "tables": [v.to_json() for v in self.tables],
            "entities": [v.to_json() for v in self.entities],
            "mappings": [v.to_json() for v in self.mappings],
            "populations": [v.to_json() for v in self.populations],
            "features": [v.to_json() for v in self.features],
            "featureSets": [v.to_json() for v in self.featureSets],
            "featureStores": [v.to_json() for v in self.featureStores],
            "destinations": [v.to_json() for v in self.destinations]
        }



"""Generated implementation of label."""

# WARNING DO NOT EDIT
# This code was generated from label.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass


@dataclasses.dataclass(frozen=True)
class Label:
    """Label metadata to apply to features and other objects.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Label data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Label:
        """Validate and parse JSON data into an instance of Label.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Label.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Label(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing Label", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> Label:
        """Parse a JSON string such as a dictionary key."""
        return Label(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)



"""Generated implementation of schedule."""

# WARNING DO NOT EDIT
# This code was generated from schedule.mcn

from __future__ import annotations

import abc  # noqa: F401
import dataclasses  # noqa: F401
import datetime  # noqa: F401
import enum  # noqa: F401
import isodate  # noqa: F401
import json  # noqa: F401
import jsonschema  # noqa: F401
import logging  # noqa: F401
import typing  # noqa: F401
import uuid  # noqa: F401
try:
    from anaml_client.utils.serialisation import JsonObject  # noqa: F401
except ImportError:
    pass


@dataclasses.dataclass(frozen=True)
class Duration:
    """A duration in ISO 8601 duration format.
    
    The "T" to separate date and time components must always be included,
    even when the string would be unambiguous without it.
    
    Example: "PT15M" for 15 minutes.
    
    Args:
        value (str): A data field.
    """
    
    value: str
    
    def __str__(self) -> str:
        """Return a str of the wrapped value."""
        return str(self.value)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for Duration data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "string"
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Duration:
        """Validate and parse JSON data into an instance of Duration.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Duration.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return Duration(str(data))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing Duration", exc_info=ex)
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return str(self.value)
    
    @classmethod
    def from_json_key(cls, data: str) -> Duration:
        """Parse a JSON string such as a dictionary key."""
        return Duration(str(data))
    
    def to_json_key(self) -> str:
        """Serialise as a JSON string suitable for use as a dictionary key."""
        return str(self.value)


@dataclasses.dataclass(frozen=True)
class RetryPolicy(abc.ABC):
    """Retry policy for jobs that fail."""
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    @classmethod
    def json_schema(cls) -> RetryPolicy:
        """JSON schema for variant RetryPolicy.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> RetryPolicy:
        """Validate and parse JSON data into an instance of RetryPolicy.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of RetryPolicy.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing RetryPolicy", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class FixedRetryPolicy(RetryPolicy):
    """Retry a failed job after a fixed backoff.
    
    Args:
        backoff (Duration): A data field.
        maxAttempts (int): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "fixed"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    backoff: Duration
    maxAttempts: int
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for FixedRetryPolicy data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "backoff": Duration.json_schema(),
                "maxAttempts": {
                    "type": "integer"
                }
            },
            "required": [
                "adt_type",
                "backoff",
                "maxAttempts",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> FixedRetryPolicy:
        """Validate and parse JSON data into an instance of FixedRetryPolicy.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of FixedRetryPolicy.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return FixedRetryPolicy(
                backoff=Duration.from_json(data["backoff"]),
                maxAttempts=int(data["maxAttempts"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing FixedRetryPolicy",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "backoff": self.backoff.to_json(),
            "maxAttempts": int(self.maxAttempts)
        }


@dataclasses.dataclass(frozen=True)
class NeverRetryPolicy(RetryPolicy):
    """Never retry a failed job."""
    
    ADT_TYPE: typing.ClassVar[str] = "never"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for NeverRetryPolicy data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> NeverRetryPolicy:
        """Validate and parse JSON data into an instance of NeverRetryPolicy.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of NeverRetryPolicy.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return NeverRetryPolicy(
                
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing NeverRetryPolicy",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE
        }


@dataclasses.dataclass(frozen=True)
class Schedule(abc.ABC):
    """Schedule specification for a job."""
    
    ADT_TYPE: typing.ClassVar[str] = ""
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    @classmethod
    def json_schema(cls) -> Schedule:
        """JSON schema for variant Schedule.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        adt_types = [klass.ADT_TYPE for klass in cls.__subclasses__()]
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": adt_types
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> Schedule:
        """Validate and parse JSON data into an instance of Schedule.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of Schedule.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            adt_type = data.get("adt_type", None)
            for klass in cls.__subclasses__():
                if klass.ADT_TYPE == adt_type:
                    return klass.from_json(data)
            raise ValueError("Unknown adt_type: '{ty}'".format(ty=adt_type))
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug("Invalid JSON data received while parsing Schedule", exc_info=ex)
            raise
    
    @abc.abstractmethod
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        raise NotImplementedError


@dataclasses.dataclass(frozen=True)
class DailySchedule(Schedule):
    """Execute the job daily.
    
    Args:
        startTimeOfDay (typing.Optional[str]): A data field.
        retryPolicy (RetryPolicy): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "daily"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    startTimeOfDay: typing.Optional[str]
    retryPolicy: RetryPolicy
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for DailySchedule data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "startTimeOfDay": {
                    "oneOf": [
                        {"type": "null"},
                        {"type": "string"},
                    ]
                },
                "retryPolicy": RetryPolicy.json_schema()
            },
            "required": [
                "adt_type",
                "retryPolicy",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> DailySchedule:
        """Validate and parse JSON data into an instance of DailySchedule.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of DailySchedule.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return DailySchedule(
                startTimeOfDay=(lambda v: v and str(v))(data.get("startTimeOfDay", None)),
                retryPolicy=RetryPolicy.from_json(data["retryPolicy"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing DailySchedule",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "startTimeOfDay": (lambda v: v and str(v))(self.startTimeOfDay),
            "retryPolicy": self.retryPolicy.to_json()
        }


@dataclasses.dataclass(frozen=True)
class CronSchedule(Schedule):
    """Execute the job according to a cron schedule string.
    
    Args:
        cronString (str): A data field.
        retryPolicy (RetryPolicy): A data field.
    """
    
    ADT_TYPE: typing.ClassVar[str] = "cron"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    cronString: str
    retryPolicy: RetryPolicy
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for CronSchedule data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                },
                "cronString": {
                    "type": "string"
                },
                "retryPolicy": RetryPolicy.json_schema()
            },
            "required": [
                "adt_type",
                "cronString",
                "retryPolicy",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> CronSchedule:
        """Validate and parse JSON data into an instance of CronSchedule.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of CronSchedule.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return CronSchedule(
                cronString=str(data["cronString"]),
                retryPolicy=RetryPolicy.from_json(data["retryPolicy"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing CronSchedule",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE,
            "cronString": str(self.cronString),
            "retryPolicy": self.retryPolicy.to_json()
        }


@dataclasses.dataclass(frozen=True)
class NeverSchedule(Schedule):
    """Do not execute the job automatically."""
    
    ADT_TYPE: typing.ClassVar[str] = "never"
    adt_type: str = dataclasses.field(init=False, repr=False, default=ADT_TYPE)
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for NeverSchedule data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "adt_type": {
                    "type": "string",
                    "enum": [cls.ADT_TYPE]
                }
            },
            "required": [
                "adt_type",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> NeverSchedule:
        """Validate and parse JSON data into an instance of NeverSchedule.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of NeverSchedule.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return NeverSchedule(
                
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing NeverSchedule",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "adt_type": self.ADT_TYPE
        }


@dataclasses.dataclass(frozen=True)
class ScheduleState:
    """Current scheduling state of a job execution.
    
    Args:
        schedule (Schedule): A data field.
        scheduledStartTime (datetime.datetime): A data field.
        retryCount (int): A data field.
    """
    
    schedule: Schedule
    scheduledStartTime: datetime.datetime
    retryCount: int
    
    @classmethod
    def json_schema(cls) -> dict:
        """Return the JSON schema for ScheduleState data.
        
        Returns:
            A Python dictionary describing the JSON schema.
        """
        return {
            "type": "object",
            "properties": {
                "schedule": Schedule.json_schema(),
                "scheduledStartTime": {
                    "type": "string",
                    "format": "date-time"
                },
                "retryCount": {
                    "type": "integer"
                }
            },
            "required": [
                "schedule",
                "scheduledStartTime",
                "retryCount",
            ]
        }
    
    @classmethod
    def from_json(cls, data: dict) -> ScheduleState:
        """Validate and parse JSON data into an instance of ScheduleState.
        
        Args:
            data (dict): JSON data to validate and parse.
        
        Returns:
            An instance of ScheduleState.
        
        Raises:
            ValidationError: When schema validation fails.
            KeyError: When a required field is missing from the JSON.
        """
        try:
            jsonschema.validate(data, cls.json_schema())
            return ScheduleState(
                schedule=Schedule.from_json(data["schedule"]),
                scheduledStartTime=isodate.parse_datetime(data["scheduledStartTime"]),
                retryCount=int(data["retryCount"]),
            )
        except jsonschema.exceptions.ValidationError as ex:
            logging.debug(
                "Invalid JSON data received while parsing ScheduleState",
                exc_info=ex
            )
            raise
    
    def to_json(self) -> dict:
        """Serialise this instance as JSON.
        
        Returns:
            Data ready to serialise as JSON.
        """
        return {
            "schedule": self.schedule.to_json(),
            "scheduledStartTime": self.scheduledStartTime.strftime('%Y-%m-%dT%H:%M:%S.%f%z'),
            "retryCount": int(self.retryCount)
        }
